//*** GAME SERVER CODE *** \\

// Node.js requirements
var util = require("util"),								// utility resources (logging, object inspection, etc)
	io = require("socket.io"),							// Socket.IO
	events = require("events"),							// events
	Player = require("./Player").Player,				// Player class
	Projectile = require("./Projectile").Projectile,	// Projectile class
	Asteroid = require("./Asteroid").Asteroid;			// Asteroid class

// core game variables	
var socket,					// socket controller
	players,				// array of connected players
	projectiles,			// array of projectiles
	asteroids;				// array of asteroids				
	
// initialization function	
function init() {
	// set the players variable to an empty array
	players = [];
	// set the projectiles variable to an empty array
	projectiles = [];
	// set the asteroids variable to an empty array
	asteroids = [];
	// get the socket server listening on a port (8000):
	socket = io.listen(8000);
	// limit Socket.IO to using WebSockets (and not falling back to anything else).
	// cut down on the volume of output Socket.IO sends to the terminal.
	socket.configure(function() {
		socket.set("transports", ["websocket"]);
		socket.set("log level", 2);
	});
	
	// start listening for events
	setEventHandlers();	
};

//*** GAME EVENT HANDLERS ***//

// listen for new connections to Socket.IO, and pass them off to onSocketConnection
var setEventHandlers = function() {
	socket.sockets.on("connection", onSocketConnection);
};

// handle Socket.IO connections
function onSocketConnection(client) {
	// listen for new socket connections
	util.log("New player has connected: "+client.id);
	// listen for client disconnects
	client.on("disconnect", onClientDisconnect);
	// listen for new player messages
	client.on("new player", onNewPlayer);
	// listen for move player messages
	client.on("move player", onMovePlayer);
	// listen for player 'fire' messages
	client.on("fire player", onFirePlayer);
	// listen for player 'create asteroid' messages
	client.on("create asteroid", onCreateAsteroid);
	// listen for projectile collision messages
	client.on("projectile collision", onProjectileCollision);
	// listen for local player collision messages
	client.on("player collision", onPlayerCollision);
	// listen for friendly fire collision messages
	client.on("friendly fire collision", onFriendlyFire);
	// listen for projectile off screen messages
	client.on("projectile off screen", onProjectileOffScreen);
	// listen for asteroid off screen messages
	client.on("asteroid off screen", onAsteroidOffScreen);
};

// called when an asteroid travels off screen,
// i.e. its Y value is > 1000:
function onAsteroidOffScreen(data) {
	// read the index of the asteroid involved
	var asteroidIndex;
	asteroidIndex = data.asteroidIndex;
	
	// remove the off-screen asteroid from the asteroids array
	asteroids.splice(asteroidIndex, 1);
	this.emit("remove asteroid", {index: asteroidIndex});
	this.broadcast.emit("remove asteroid", {index: asteroidIndex});
};

// called when a projectile travels off screen,
// i.e. its Y value is < 0:
function onProjectileOffScreen(data) {
	// read the index of the projectile involved
	var projectileIndex;
	projectileIndex = data.projectileIndex;
	
	// remove the off-screen projectile from the projectiles array
	projectiles.splice(projectileIndex, 1);
	this.emit("remove projectile", {index: projectileIndex});
	this.broadcast.emit("remove projectile", {index: projectileIndex});
};

// called when a collision between a projectile and a player occurs.
// this is a friendly fire death!
function onFriendlyFire(data) {
	// read the index of the projectile involved
	var projectileIndex;
	projectileIndex = data.projectileIndex;
	
	// remove the destroyed projectile from the projectiles array
	// (for now, every collision results in instant death).
	projectiles.splice(projectileIndex, 1);
	this.emit("remove projectile", {index: projectileIndex});
	this.broadcast.emit("remove projectile", {index: projectileIndex});
	
	// find the dead player using his id number
	var removePlayer = playerById(this.id);
	
	if (!removePlayer) {
		util.log("Player not found: "+this.id);
		return;
	};
	
	// remove the dead player from the players array
	players.splice(players.indexOf(removePlayer), 1);
	this.broadcast.emit("remove player", {id: this.id});
	this.emit("you're dead friendly fire");
};

// called when a collision between an asteroid and a player occurs
function onPlayerCollision() {
	// find the dead player using his id number
	var removePlayer = playerById(this.id);
	
	if (!removePlayer) {
		util.log("Player not found: "+this.id);
		return;
	};
	
	// remove the dead player from the players array
	players.splice(players.indexOf(removePlayer), 1);
	this.broadcast.emit("remove player", {id: this.id});
	this.emit("you're dead");
};

// called when a collision between an asteroid and a projectile occurs
function onProjectileCollision(data) {
	// read the indices of the asteroid and projectile involved
	var asteroidIndex, projectileIndex;
	asteroidIndex = data.asteroidIndex;
	projectileIndex = data.projectileIndex;
	
	// remove the destroyed asteroid from the asteroids array
	// (for now, every collsion results in instant death).
	asteroids.splice(asteroidIndex, 1);
	this.emit("remove asteroid", {index: asteroidIndex});
	this.broadcast.emit("remove asteroid", {index: asteroidIndex});
	
	// remove the destroyed projectile from the projectiles array
	// (for now, every collision results in instant death).
	projectiles.splice(projectileIndex, 1);
	this.emit("remove projectile", {index: projectileIndex});
	this.broadcast.emit("remove projectile", {index: projectileIndex});
};

// called when clients prompt to create asteroids
// this happens on a timed loop, for each connected client.
function onCreateAsteroid() {
	// calculate a random start position for the new asteroid.
	// the minus 10 (half the maximum asteroid size - 20) stops the asteroid from being placed 
	// right on the egde of the screen.
	// replaced canvas.width with 900,
	// replaced canvas.height with 450 since these can vary from one player to another.
	
	var startX = Math.round(Math.random()*(900-3)),
		startY = Math.round(Math.random()*(450-3));
	// set the speed of the asteroid - the xSpeed can be between -5 and 5, y can be between 2 and 5
	var xLeftorRight = Math.round(Math.random());
	if (xLeftorRight == 0) {
		xSpeed = Math.round(Math.random()*(5));
	} else {
		xSpeed = (-1) * Math.round(Math.random()*(5));
	}
	ySpeed = Math.floor((Math.random()*5)+2);
		
	// select a random number between 0 and 4.
	// asteroids use this number to decide which shape they will take on.
	// we have 5 asteroid images, named a0, a1, a2, a3, a4.
	// the aNum will tell the Asteroid class which image to use to draw.
	var aNum = Math.round(Math.random()*(4));
	this.emit("new asteroid", {x: startX, y: startY, xSpeed: xSpeed, ySpeed: ySpeed, aNum: aNum});
	this.broadcast.emit("new asteroid", {x: startX, y: startY, xSpeed: xSpeed, ySpeed: ySpeed, aNum: aNum});
};

// called when a player disconnects from the game
function onClientDisconnect() {
	util.log("Player has disconnected: "+this.id);
	
	// find the disconnected player using his id number
	var removePlayer = playerById(this.id);
	
	if (!removePlayer) {
		util.log("Player not found: "+this.id);
		return;
	};
	
	// remove the disconnected player from the players array
	players.splice(players.indexOf(removePlayer), 1);
	this.broadcast.emit("remove player", {id: this.id});
};

// called when a new player joins the game.
// create a new player instance using position data sent by the connected client,
// and store the identification number for future reference.
function onNewPlayer(data) {
	var newPlayer = new Player(data.x, data.y);
	newPlayer.id = this.id;
	
	// notify other players of the new player
	this.broadcast.emit("new player", {id: newPlayer.id, x: newPlayer.getX(), y: newPlayer.getY()});
	
	// send existing players to the new player
	var i, existingPlayer;
	for (i = 0; i < players.length; i++) {
		existingPlayer = players[i];
		this.emit("new player", {id: existingPlayer.id, x: existingPlayer.getX(), y: existingPlayer.getY()});
	};
	
	// send existing asteroids to the new player
	var j, nextAsteroid;
	for (j = 0; j < asteroids.length; j++) {
		nextAsteroid = asteroids[j];
		this.emit("new asteroid", {x: nextAsteroid.getX(), y: nextAsteroid.getY(), xSpeed: nextAsteroid.getxSpeed(), ySpeed: nextAsteroid.getySpeed(), size: nextAsteroid.getSize()});
	};
	
	// add the new player to the players array
	players.push(newPlayer);	
};

// called when a player moves
function onMovePlayer(data) {
	// search for the player that is being moved
	var movePlayer = playerById(this.id);

	if (!movePlayer) {
		util.log("Player not found: "+this.id);
		return;
	};
	
	// update the player's x and y position
	movePlayer.setX(data.x);
	movePlayer.setY(data.y);
	
	// broadcast the player's updated position to the other players
	this.broadcast.emit("move player", {id: movePlayer.id, x: movePlayer.getX(), y: movePlayer.getY()});
};

// called when a player fires a weapon
function onFirePlayer(data) {
	// we need the x, y coords of the firing player to select the initial position
	var newProjectile = new Projectile(data.x, data.y);

	// notify the firing player of the new projectile
	this.broadcast.emit("fire player", {x: newProjectile.getX(), y: newProjectile.getY()});
	// broadcast the new projectile to the other players
	this.emit("fire player", {x: newProjectile.getX(), y: newProjectile.getY()});
	
	// add the new projectile to the projectiles array
	projectiles.push(newProjectile);
};

// helper function to locate player within the players array, using supplied id:
function playerById(id) {
	var i;
	for (i = 0; i < players.length; i++) {
		if (players[i].id == id)
			return players[i];
	};
	
	return false;
};

// run the game!
init();

// *** =============================================================================================
// *** (commented out, not-fully-working bits below)

// commented out b/c the client code wasn't working right.
// might be something to add later, but not really necessary.
/*
// called when a client reports another player (a remote player)
// has collided with an asteroid
function onRemotePlayerCollision(data) {
	util.log("remote player collision message received!");
	util.log("dead player ID is: "+data.deadPlayerID);
	
	// figure out who the dead player is by his ID:
	var deadPlayer = playerById(data.deadPlayerID);
	
	if (!deadPlayer) {
		util.log("Player not found: "+data.deadPlayerID);
		//return;
	};
	
	// remove the dead player from the players array
	players.splice(players.indexOf(deadPlayer), 1);
	this.broadcast.emit("remove player", {id: data.deadPlayerID});
	this.emit("remove player", {id: data.deadPlayerID});
	//this.emit("you're dead");
	//deadPlayer.emit("you're dead");
	//socket.sockets.socket(data.deadplayerID).emit("you're dead");
};
*/
